
--[[

Miscellaneous functions, including:

- A simple math.round type function (avoids string conversion)
- A series of value-to-string type formatting functions
- A series of color manipulation functions
- A simple one-time timer class, with a creation order guarantee

]]--

_G.Misc = {}

-- MATH UTILS

function Misc.Round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

-- FORMATTING

-- NB: assumes a number equal or greater than 1,000
local function FormatNumber(number)
	number = Misc.Round(number);
	
	-- if the number is below 100,000 display it in full
	if (number < 100000) then
		return string.format("%d,%03d",number/1000,number%1000);
	end
	
	-- otherwise, for very large numbers, round them further (depending on the value)
	number = number/1000;
	local thousands = 1;
	while (number > 1000) do
		number = number/1000;
		thousands = thousands + 1;
		if (thousands > 3) then break end
	end
	
	return string.format("%#."..
		(number < 10 and 3 or (number < 100 and 2 or (number < 1000 and 1 or 0))).."f"..
		(thousands == 1 and L.Thousand or (thousands == 2 and L.Million or (thousands == 3 and L.Billion or L.Trillion))),
			number
	);
end

-- format a standard value
function Misc.FormatValue(value)
	local number = (value+0);
	
	-- if the number is below 10,000 display the full number with (up to) one decimal place
	value = Misc.Round(number,1);
	if (value < 1000) then
		return value;
	elseif (value < 10000) then
		return string.format("%d,%"..(value%1 == 0 and "03d" or "05.1f"),value/1000,value%1000);
	end
	
	-- display the number in the standard long format
	return FormatNumber(number);
end

-- format a "per second" value
function Misc.FormatPs(ps)
	local number = (ps+0);
	
	-- if the number is less than 10,000 display the number (always) rounded to one decimal
	ps = Misc.Round(number,1);
	if (ps < 1000) then
		return string.format("%#.1f",ps);
	elseif (ps < 10000) then
		return string.format("%d,%05.1f",ps/1000,ps%1000);
	end
	
	-- display the number in the standard long format
	return FormatNumber(ps);
end

-- format a "percentage"
function Misc.FormatPerc(perc,round100)
	perc = 100*perc;
	
	--  if the number is below 10,000%, we always round to one decimal place (unless specified not to when the value is exactly 100%)
	if (perc < 1000) then
		return (((perc+0) == 100 and round100) and "100%" or string.format("%.1f%%",perc));
	elseif (perc < 10000) then
		return string.format("%d,%03.1f%%",perc/1000,perc%1000);
	end
	
	-- display the number in standard long format followed by a % sign (TODO: will need to consider this case further at some point)
	return FormatNumber(perc).."%";
end

-- format a time (format = [hh]:[mm]:[ss][am/pm])
function Misc.FormatTime(lotroDate)
	-- am/pm
	if (locale == "en") then
		return string.format("%02d",(lotroDate.Hour == 12 and 12 or lotroDate.Hour%12))..":"..string.format("%02d",lotroDate.Minute)..":"..string.format("%02d",lotroDate.Second)..(lotroDate.Hour>=12 and "pm" or "am");
	-- 24h
	else
		return string.format("%02d",lotroDate.Hour)..":"..string.format("%02d",lotroDate.Minute)..":"..string.format("%02d",lotroDate.Second);
	end
end

local function NumberToMonth(number,short)
	if (number == 1)  then return (short and L.Jan[1] or L.Jan[2]) end
	if (number == 2)  then return (short and L.Feb[1] or L.Feb[2]) end
	if (number == 3)  then return (short and L.Mar[1] or L.Mar[2]) end
	if (number == 4)  then return (short and L.Apr[1] or L.Apr[2]) end
	if (number == 5)  then return (short and L.May[1] or L.May[2]) end
	if (number == 6)  then return (short and L.Jun[1] or L.Jun[2]) end
	if (number == 7)  then return (short and L.Jul[1] or L.Jul[2]) end
	if (number == 8)  then return (short and L.Aug[1] or L.Aug[2]) end
	if (number == 9)  then return (short and L.Sep[1] or L.Sep[2]) end
	if (number == 10) then return (short and L.Oct[1] or L.Oct[2]) end
	if (number == 11) then return (short and L.Nov[1] or L.Nov[2]) end
	if (number == 12) then return (short and L.Dec[1] or L.Dec[2]) end
end

-- format a date (format = dd mmm yyyy)
function Misc.FormatDate(lotroDate)
	return string.format("%02d",lotroDate.Day).." "..NumberToMonth(lotroDate.Month,true).." "..string.format("%04d",lotroDate.Year);
end

-- format a duration (format = [m]m [ss]s) - if slim is true, and the duration is less than 1m, don't show the minutes
function Misc.FormatDuration(duration,slim)
	if (slim and duration/60 < 1) then
		return string.format("%.1f",duration%60)..L.SecondsAbbr;
	else
		return string.format("%d",duration/60)..L.MinutesAbbr.." "..string.format("%.1f",duration%60)..L.SecondsAbbr;
	end
end

-- determine which time is the minimum
function Misc.MinTime(time1,time2)
	if (time1.Year ~= time2.Year) then return (time1.Year < time2.Year and time1 or time2) end
	if (time1.Month ~= time2.Month) then return (time1.Month < time2.Month and time1 or time2) end
	if (time1.Day ~= time2.Day) then return (time1.Day < time2.Day and time1 or time2) end
	if (time1.Hour ~= time2.Hour) then return (time1.Hour < time2.Hour and time1 or time2) end
	if (time1.Minute ~= time2.Minute) then return (time1.Minute < time2.Minute and time1 or time2) end
	return (time1.Second < time2.Second and time1 or time2);
end

-- SORTING

function Misc.SortData(unsortedData,sortOnDurations)
	local sortedData = {};
	-- extract data
	for playerName,data in pairs(unsortedData) do
		if not data.empty then
			table.insert(sortedData,{playerName,data});
		end
	end
	-- sort data
	if (sortOnDurations) then
		Misc.SortOnDurations(sortedData);
	else
		Misc.Sort(sortedData);
	end
	-- return sorted data
	return sortedData;
end

function Misc.Sort(unsortedData)
	table.sort(unsortedData,function(a,b)
		if (a[2]:TotalAmount() > b[2]:TotalAmount()) then
			return true;
		elseif (a[2]:TotalAmount() == b[2]:TotalAmount()) then
			if (a[2].attacks > b[2].attacks) then
				return true;
			elseif (a[2].attacks == b[2].attacks) then
				return string.lower(a[1]) < string.lower(b[1]);
			else
				return false;
			end
		else
			return false;
		end
	end);
end

function Misc.SortOnDurations(unsortedData)
	local timestamp = Turbine.Engine.GetGameTime();
	table.sort(unsortedData,function(a,b)
		if (a[2]:CurrentDuration(timestamp) > b[2]:CurrentDuration(timestamp)) then
			return true;
		elseif (a[2]:CurrentDuration(timestamp) == b[2]:CurrentDuration(timestamp)) then
			if (a[2].applications == nil or (b[2].applications ~= nil and a[2].applications > b[2].applications)) then
				return true;
			elseif (b[2].applications ~= nil and a[2].applications == b[2].applications) then
				return string.lower(a[1]) < string.lower(b[1]);
			else
				return false;
			end
		else
			return false;
		end
	end);
end

-- EQUALITY CHECK

function Misc.TablesEqual(table1,table2)
	-- if the supplied arguments are not tables, check for equality
	if (type(table1) ~= "table") then
		return table1 == table2;
		
	-- if the supplied arguments are tables, check each element for equality
	elseif (type(table1) == "table") then
		if (type(table2) ~= "table") then return false end
		
		for k,v in pairs(table1) do
			if (not Misc.TablesEqual(v,table2[k])) then return false end
		end
		for k,v in pairs(table2) do
			if (table1[k] == nil) then return false end
		end
		
		return true;
	end
end

-- DEEP TABLE COPY

function Misc.TableCopy(originalTable,newTable)
	if (type(originalTable) ~= "table") then return originalTable end
	
	if (newTable == nil) then newTable = {} end
	
	for k,v in pairs(originalTable) do
		newTable[Misc.TableCopy(k)] = Misc.TableCopy(v);
	end
	return newTable;
end

-- COLOR MANIPULATION

function Misc.ToGray(alpha,percent,color,strength)
	local maxColor = math.max(color.R,color.G,color.B);
	local newRed,newGreen,newBlue;
	if maxColor == color.R then
		newRed = color.R;
		newGreen = (color.R-(1-strength)*(color.R-color.G))*(1-percent*(color.R-color.G))
		newBlue = (color.R-(1-strength)*(color.R-color.B))*(1-percent*(color.R-color.B))
	elseif maxColor == color.G then
		newRed = (color.G-(1-strength)*(color.G-color.R))*(1-percent*(color.G-color.R))
		newGreen = color.G;
		newBlue = (color.G-(1-strength)*(color.G-color.B))*(1-percent*(color.G-color.B))
	else
		newRed = (color.B-(1-strength)*(color.B-color.G))*(1-percent*(color.B-color.R))
		newGreen = (color.B-(1-strength)*(color.B-color.G))*(1-percent*(color.B-color.G))
		newBlue = color.B;
	end
	return Turbine.UI.Color(alpha,newRed,newGreen,newBlue);
end

function Misc.Shade(color,amount)
	return Turbine.UI.Color(color.A,color.R-(color.R)*(0.1+0.05*amount),color.G-(color.G)*(0.1+0.05*amount),color.B-(color.B)*(0.1+0.05*amount));
end

function Misc.SetAlpha(color,alpha)
	return Turbine.UI.Color(alpha,color.R,color.G,color.B);
end

local ToHex = function(base16Number)
	if (base16Number < 10) then 		 return string.format("%0d",base16Number);
	elseif (base16Number < 11) then  return "A";
	elseif (base16Number < 12) then  return "B";
	elseif (base16Number < 13) then  return "C";
	elseif (base16Number < 14) then  return "D";
	elseif (base16Number < 15) then  return "E";
	else 									  				 return "F";
	end
end

function Misc.DecimalToHex(decimal)
	return "#"..
		ToHex((decimal.R*255)/16)..ToHex((decimal.R*255)%16)..
		ToHex((decimal.G*255)/16)..ToHex((decimal.G*255)%16)..
		ToHex((decimal.B*255)/16)..ToHex((decimal.B*255)%16);
end

-- LISTENERS (listeners are assigned to any kind of object)

Misc.listeners = {}

function Misc.AddListener(object,listenerFunction,listenerObject)
	if (Misc.listeners[object] == nil) then
		Misc.listeners[object] = {{listenerFunction,listenerObject}};
	else
		table.insert(Misc.listeners[object],{listenerFunction,listenerObject});
	end
end

function Misc.NotifyListeners(object)
	if (Misc.listeners[object] ~= nil) then
		for _,listener in ipairs(Misc.listeners[object]) do
			listener[1](listener[2]);
		end
	end
end

function Misc.RemoveListener(object,listenerFunction,listenerObject)
	if (Misc.listeners[object] == nil) then return end
	
	for i,l in ipairs(Misc.listeners[object]) do
		if (listenerFunction == l[1] and listenerObject == l[2]) then
			table.remove(Misc.listeners,i);
			return;
		end
	end
end

-- shortcut method to update all listeners when a global variable is changed
function Misc.SetValue(variableName,value)
	_G[variableName] = value;
	Misc.NotifyListeners(variableName);
end

-- SIMPLE TIMER

-- note we do guarantee that timer A, who is started before (but with the
--  same timestamp as) and has the same duration as timer B, will always
--  call its listener (notify) function first
-- 
-- a listener object is not required if the supplied function is static

local timerControl = Turbine.UI.Control();
timerControl.timers = {}

timerControl.Update = function()
	local gameTime = Turbine.Engine.GetGameTime();
	local indexesToRemove = {}
	-- notify listeners of any completed timers (in order), and remove them from the list
	local index = 1;
	while true do
		if (index > #timerControl.timers) then break end
		
		local timerData = timerControl.timers[index];
		if (gameTime >= timerData[1]) then
			table.remove(timerControl.timers,index);
			if (timerData[3]) then
				timerData[2](timerData[3],timerData[4],timerData[5]);
			else
				timerData[2](timerData[4],timerData[5]);
			end
		else
			index = index+1;
		end
	end
	-- if no timers remain, stop checking for finished timers
	if (#timerControl.timers == 0) then
		timerControl:SetWantsUpdates(false);
	end
end

function Misc.StartTimer(target,timestamp,duration,listenerFunction,listener)
	-- if this is the first timer added, start checking for it to finish
	if (#timerControl.timers == 0) then
		timerControl:SetWantsUpdates(true);
	end
	-- create the timer
	table.insert(timerControl.timers,{timestamp+duration,listenerFunction,listener,timestamp,target});
end

-- for compatability with other plugins, this method is used to not override the same events

function Misc.AddCallback(object, event, callback)
	if (object[event] == nil) then
		object[event] = callback;
	else
		if (type(object[event]) == "table") then
			table.insert(object[event], callback);
		else
			object[event] = {object[event], callback};
		end
	end
end

--- DETERMINE TEXT LENGTH

window = Turbine.UI.Window();

local label = Turbine.UI.Label();
label:SetParent(window);
label:SetHeight(25);
label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
label:SetMultiline(false);

local scroll = Turbine.UI.Lotro.ScrollBar();
scroll:SetOrientation(Turbine.UI.Orientation.Horizontal);
scroll:SetHeight(10);
scroll:SetParent(window);
scroll:SetSize(60,10);
label:SetHorizontalScrollBar(scroll);

window:SetPosition(-100,-100);
window:SetSize(100,100);
window:SetVisible(true);
window:SetWantsUpdates(true);

Misc.texts = {}

window.Update = function(sender)
	for text,font in pairs(Misc.texts) do
		label:SetFont(font);
		label:SetText(text);
		
		local width = 0;
		local increment = 512;
		local count = 0;
		
		while true do
			label:SetWidth(width);
			if (not scroll:IsVisible()) then
				if (increment < 5) then
					Misc.texts[text] = width;
					break;
				end
				width = width - increment;
				increment = increment/2;
			end
			width = width + increment;
		end
	end
	
	sender:SetWantsUpdates(false);
	sender:SetVisible(false);
	sender = nil;
end

function Misc.DetermineLength(text,font)
	Misc.texts[text] = font;
end

for i=1,100 do
	Misc.DetermineLength("A really really really long test with a number on the end just to make sure that this doesnt take too long, hnmm: "..i,Turbine.UI.Lotro.Font.TrajanPro14);
end

